/*
Miguel Angel Armas.
clase del 14 de Febrero de 1994.
Programacin Avanzada (C++).
Fundacin Arturo Rosenblueth.
//------------------------------------------------------------------------
  class (keyword)
 
 Defines C++ class

 Syntax:  <classkey> <classname> [<:baselist>] { <member list> }

Classes are specific to C++.
   <classkey> is one of the keywords "class", "struct", or "union".
   <classname> can be any name unique within its scope.
   <baselist> lists the base class(es) that this class derives from.
    <baselist> is optional.
   <member list> declares the class's data members and member functions.

Within a class,
   the data are called "data members"
   the functions are called "member functions"

//------------------------------------------------------------------------

  public, private, and protected (access specifiers)
 
Members of a class can acquire access attributes in one of two ways:
   by default, or
   through the use of the access specifiers
    public, private, and protected.

 Syntax:
   public: <declarations>
   private: <declarations>
   protected: <declarations>

 public
 
If a member is public, it can be used by any function.

In C++, members of a struct or union are public by default.

 private
 
If a member is private, it can only be used by member functions and friends
of the class in which it is declared.

Members of a class are private by default.

 protected
 
If a member is protected, its access is the same as for private.

In addition, the member can be used by member functions and friends of
classes derived from the declared class, but only in objects of the derived
type.


You can override the default struct access with private or protected. You
can't override the default union access.

Friend declarations are not affected by these access specifiers.

//------------------------------------------------------------------------

  Virtual Classes
 
You might want to make a class virtual if it is a base class that has been
passed to more than one derived class, as might happen with multiple
inheritance.

A base class can't be specified more than once in a derived class:

  class B { ...};
  class D : B, B { ... };  // ILLEGAL

However, a base class can be indirectly passed to the derived class more
than once:

  class X : public B { ... }
  class Y : public B { ... }
  class Z : public X, public Y { ... }  // OK

In this case, each object of class Z will have two sub-objects of class B.

If this causes problems, you can add the keyword "virtual" to a base class
specifier. For example,

  class X : virtual public B { ... }
  class Y : virtual public B { ... }
  class Z : public X, public Y { ... }

B is now a virtual base class, and class Z has only one sub-object of class
B.


 Constructors for Virtual Base Classes
 
Constructors for virtual base classes are invoked before any non-virtual
base classes.

If the hierarchy contains multiple virtual base classes, the virtual base
class constructors are invoked in the order in which they were declared.

Any non-virtual bases are then constructed before the derived class
constructor is called.

If a virtual class is derived from a non-virtual base, that non-virtual base
will be first, so that the virtual base class can be properly constructed.
For example, this code

  class X : public Y, virtual public Z
     X one;

produces this order:

  Z();   // virtual base class initialization
  Y();   // non-virtual base class
  X();   // derived class

//------------------------------------------------------------------------

  Virtual functions
 
Virtual functions let derived classes provide different versions of a base
class function.

You can declare a virtual function in a base class, then redefine it in any
derived class, even if the number and type of arguments are the same.

The redefined function overrides the base class function of the same name.
Virtual functions can only be member functions.

You can also declare the functions

  int Base::Fun(int)

and

  int Derived::Fun(int)

even when they are not virtual.

The base class version is available to derived class objects via scope
override.

If they are virtual, only the function associated with the actual type of
the object is available.

With virtual functions, you can't change just the function type. It is
illegal, therefore, to redefine a virtual function so that it differs only
in the return type.

If two functions with the same name have different arguments, C++ considers
them different, and the virtual function mechanism is ignored.

//------------------------------------------------------------------------

  friend (keyword)
 
A friend of a class X is a function or class that, although not a member of
that class, has full access rights to the private and protected members of
class X.

 Syntax:  friend <identifier>;

In all other respects, the friend is a normal function in terms of scope,
declarations, and definitions.

 Example:
  class stars {
    friend galaxy;
    int magnitude;
    int starfunc(void);
  };
  class galaxy {
    long int number_of_stars;
    void stars_magnitude(stars&);
    void stars_func(stars*);
  }

//------------------------------------------------------------------------

 
  Predefined C++ stream objects
 
C++ provides these predefined stream objects:

 
  cin  Standard input, usually keyboard, corresponding to stdin in C
  cout Standard output, usually screen, corresponding to stdout in C
  cerr Standard error output, usually screen, corresponding to stderr in C
  clog A fully-buffered version of cerr (no C equivalent)

You can redirect cin, cout, cerr, and clog from and to other devices and
files.

Here is a simplified view of the streamable classes, from primitive to
specialized:

  IOStream Class      What It Does
 
  streambuf           Provides methods for memory buffers
  ios                 Handles stream state variables and errors
  istream             Handles formatted and unformatted character
		      conversions from a streambuf
  ostream             Handles formatted and unformatted character
		      conversions to a streambuf
		     
  iostream            Combines istream and ostream to handle bidirectional
		      operations on a single stream
  istream_withassign  Provides constructors and assignment operators for
		      the cin stream
  ostream_withassign  Provides constructors and assignment operators for
		      cout, cerr and clog

//------------------------------------------------------------------------

  new (operator) and delete (operator)
 
 Operators that create and destroy an object

 Syntax:
   <pointer_to_name> = new <name> [ <name_initializer> ];
   delete <pointer_to_name>;

The "new" operator tries to create an object <name> by allocating
sizeof(<name>) bytes in the heap.

The "delete" operator destroys the object <name> by deallocating
sizeof(<name>) bytes (pointed to by <pointer_to_name>).

The storage duration of the new object is from the point of creation until
the operator "delete" deallocates its memory, or until the end of the
program.

 Example:
  name *nameptr;  // name is any non-function type
   ...
  if (!(nameptr = new name)) {
    errmsg("Insufficient memory for name");
    exit (1);
  }
  // Use *nameptr to initialize new name object
   ...
  delete nameptr; //destroy name; deallocate sizeof(name) bytes
*/
//------------------------------------------------------------------------
//-------------------Inicia Cdigo -----------------------------------------
#include <iostream.h>

class MyListData {
public:
	virtual void Print(void);
};

void MyListData::Print(void){
	cout << "No Print Method\n";
}
//---------------------------------------------------------------------------
class MyListElement{
	MyListData *data;
	MyListElement *next;
public:
	MyListElement(MyListData *initialData)
		{ data = initialData; next = NULL; }
	friend class MyList;
};
//---------------------------------------------------------------------------
class MyList : public MyListData {
	MyListElement *head;
	MyListElement *tail;
public: MyList(void) { head = tail = NULL; }
	void AddToList(MyListData *data);
	void Print(void);
};

void MyList::AddToList(MyListData *data){
	MyListElement *newElement = new MyListElement(data);
	if (head == NULL)
		head = tail = newElement;
	else{
		tail->next = newElement;
		tail = newElement;
	}
}

void MyList::Print(void){
	MyListElement *element;
	for(element = head; element!=NULL; element = element->next)
		element->data->Print();
}
//---------------------------------------------------------------------------
class MyNumber : public MyListData{
	int value;
public:
	MyNumber(int initValue) { value = initValue; }
	virtual void Print(void);
};

void MyNumber::Print(void){
	cout << "Number: " << value << "\n";
}

class MyPoint : public MyListData {
	int x;
	int y;
public:
	MyPoint(int initialX,int initialY) { x = initialX; y = initialY; }
	virtual void Print(void);
};

void MyPoint::Print(void){
	cout << "Point: " << x << "," << y << "\n";
}

void main(void){
	MyList *list1 = new MyList;
	MyList *list2 = new MyList;
	MyNumber *n1 = new MyNumber(10);
	MyNumber *n2 = new MyNumber(20);
	MyPoint *p1 = new MyPoint(2,3);
	MyPoint *p2 = new MyPoint(4,5);
	//construir las listas
	list1->AddToList(n1);
	list1->AddToList(n2);
	list1->AddToList(p1);

	list2->AddToList(n2);
	list2->AddToList(p1);
	list2->AddToList(p2);

	list2->AddToList(list1);
//	list1->AddToList(list2);

	//imprimir la lista
	cout << "list1: \n";
	list1->Print();
	cout << "list2: \n";
	list2->Print();
}